@@ -8,7 +8,7 @@ class GLP(NonLinProblem):
     expectedArgs = ['f', 'x0']
     allowedGoals = ['minimum', 'min', 'maximum', 'max']
     goal = 'minimum'
-    showGoal = False
+    showGoal = True
     isObjFunValueASingleNumber = True
     plotOnlyCurrentMinimum= True
     _currentBestPoint = None
@@ -45,4 +45,5 @@ class GLP(NonLinProblem):
                 self.lb = -inf * ones(self.n)
             if not hasattr(self, 'ub'):
                 self.ub =  inf * ones(self.n)
-            if 'x0' not in kwargs.keys(): self.x0 = (asarray(self.lb) + asarray(self.ub)) / 2.0
+            if 'x0' not in kwargs and len(args) < 2: 
+                self.x0 = (asarray(self.lb) + asarray(self.ub)) / 2.0
