@@ -15,16 +15,23 @@ class MCP(MatrixProblem):
 
     def __init__(self, *args, **kw):
         MatrixProblem.__init__(self, *args, **kw)
-        self.__init_kwargs = kw
+        self._init_kwargs = kw
         self._init = True
         
+    def manage(self, *args, **kw):
+        kw['routine'] = 'manage'
+        r = self.solve(*args, **kw)
+        r.probType = self.probType
+        return r
+        
     def solve(self, *args, **kw):
         import networkx as nx
         graph = nx.complement(self.graph)
         from openopt import STAB
-        KW = self.__init_kwargs
+        KW = self._init_kwargs
         KW.update(kw)
         P = STAB(graph, **KW)
         r = P.solve(*args)
+        r.probType = self.probType
         return r
         
