@@ -1,5 +1,6 @@
+PythonAny = any
 PythonMin, PythonMax = min, max
-from numpy import isnan, array, atleast_1d, asarray, all, searchsorted, logical_or, any, nan, \
+from numpy import isnan, array, atleast_1d, all, searchsorted, logical_or, any, nan, \
 vstack, inf, where, logical_not, min, abs, insert, logical_xor, argsort, zeros_like
 
 # for PyPy
@@ -22,94 +23,153 @@ hasPoint = lambda y, e, point:\
     True if y.size != 0 and any([(all(y[i]<=point) and all(e[i]>=point)) for i in range(y.shape[0])]) else False
 pointInd = lambda y, e, point:\
     where([(all(y[i]<=point) and all(e[i]>=point)) for i in range(y.shape[0])])[0].tolist()
-    
+
+a = np.array([-0.39])
+
+
 def r14(p, nlhc, residual, definiteRange, y, e, vv, asdf1, C, 
                  r40, g, nNodes, 
                  r41, fTol, Solutions, varTols, _in, dataType,
-                 maxNodes, _s, indTC, xRecord):
-
+                 maxNodes, _s, indTC, xRecord, activeCons):
+    assert y.shape[0] == _s.size
+#    print('!', p.iter, hasPoint(y, e, a))
     isSNLE = p.probType in ('NLSP', 'SNLE')
+#    print('iter:',p.iter,'r41:',r41)
+
+    #debug!!!!!!!!!!
+#    if len(_in) == 0 and y.size == 0:
+#        print('no points left')
+#    else:
+#        _lx, _ux = vstack([node.y for node in _in]+[y]), vstack([node.e for node in _in]+[e])
+#        if p.iter>=197: #not hasPoint(_lx, _ux, a):
+#            print('iter>>', p.iter, 'in start of preocessBoxes:', hasPoint(_lx, _ux, a))
+
     
-#    P = array([-0.63521194458007812, -0.3106536865234375, 0.0905609130859375, 0.001522064208984375, -0.69999999999999996, -0.99993896484375, 0.90000152587890625, 1.0, 4.0])
-#    p.cp = P
-#    print(0, p.iter, hasPoint(y, e, P), pointInd(y, e, P))
-#    if p.iter == 13:
-#        print y[1]
-#        print e[1]
-#        y = y[1].reshape(1, -1)
-#        e = e[1].reshape(1, -1)
-#        _s = atleast_1d([_s[1]])
-#        indTC = atleast_1d([indTC[1]])
-        
-#    if p.iter == 99:
-#        y = y[14].reshape(1, -1)
-#        e = e[14].reshape(1, -1)
+#    print p.iter, y, e, hasPoint(y, e, [10.,  20.,  10.,  50])
 
-    maxSolutions, solutions, coords = Solutions.maxNum, Solutions.solutions, Solutions.coords
-    if len(p._discreteVarsNumList):
-        y, e, _s, indTC = adjustDiscreteVarBounds(y, e, _s, indTC, p)
+    #maxSolutions, solutions, coords = Solutions.maxNum, Solutions.solutions, Solutions.coords
+    maxSolutions, solutions = Solutions.maxNum, Solutions.solutions
 
+    fo_prev = \
+    float(0 if isSNLE else PythonMin((r41, 
+    r40 - (fTol if maxSolutions == 1 else 0))))
+    fo_prev  =  PythonMin(1e300, fo_prev)
+    
     if 0 and isSNLE:
         a = residual
         o = zeros_like(a)
-        o2, a2, r41 = r45(y, e, vv, p, asdf1, dataType, r41, nlhc)
+        o2, a2, r41 = r45(y, e, vv, p, asdf1, dataType, 
+        r41, nlhc)
 #        print p.iter,nanmax(a/a2) / nanmin(a/a2)
 #        o, a = o2, a2
         a = a2
         o = zeros_like(a)
     else:
-        o, a, r41 = r45(y, e, vv, p, asdf1, dataType, r41, nlhc)
-
-#    print(o, a)
-#    print 'LF:', o
-#    print 'a:', a
-#    if p.iter == 13:
-#        print y[1]
-#        print e[1]
-#        print o[1]
-#        print a[1]
-
-#    if not isSNLE and isfinite(r40):
-#        tmp1 = r40
-#        tmp2 = PythonMin(nanmin(o), PythonMin([1e300]+[node.key for node in _in]))
-#        rTol = p.rTol * (tmp1 - tmp2) / (1e-300+PythonMin(abs(tmp1), abs(tmp2)))
-#        fTol = PythonMax(rTol, fTol)
-#        print p.iter, fTol
-
-    fo_prev = \
-    float(0 if isSNLE else PythonMin((r41, r40 - (fTol if maxSolutions == 1 else 0))))
-    fo_prev  =  PythonMin(1e300, fo_prev)
+        y, e, o, a, r41, _s, indTC = \
+        r45(y, e, vv, p, asdf1, dataType, r41, nlhc, 
+        _s, indTC)
+        #assert _s.size == y.shape[0]
+        
+#        print(p.iter, o, a)
     
-    y, e, o, a, _s, indTC, nlhc, residual = \
-    func7(y, e, o, a, _s, indTC, nlhc, residual)    
-
-    if y.size == 0:
+    assert len(activeCons) == _s.size == y.shape[0]
+    
+    if o is None:
         return _in, g, fo_prev, _s, Solutions, xRecord, r41, r40
     
-    nodes = func11(y, e, nlhc, indTC, residual, o, a, _s, p)
+    
+    # was moved to interalgLLR.func11()
+#    y, e, o, a, _s, indTC, nlhc, residual, activeCons = \
+#    func7(y, e, o, a, _s, indTC, nlhc, residual, activeCons)    
+
+    
+    
+#    if p.iter > 9:
+#        print('3>>', p.iter, hasPoint(y, e, a))
+
+#    #debug
+#    if p.iter > 116:
+#        N = _in
+#        print('len(N):', len(N))
+#        print('m:', y.shape[0])
+##        N = an
+#        active_cons = [node.activeCons for node in N] + activeCons
+#        cons_counter = p._constraints_counter
+#        c2 = dict()
+#        for _cons_list in active_cons:
+#            for elem in _cons_list:
+#                tmp = c2.get(elem, None)
+#                if tmp is None:
+#                    c2[elem] = 1
+#                else:
+##                    tmp += 1
+#                    c2[elem] += 1
+#        print(sum(c2.values()), sum(cons_counter.values()))
+#        print('-----------')
+
+    
+    nodes, y, e, o, a, _s, indT, nlhc, residual, activeCons = func11(y, e, nlhc, indTC, residual, o, a, _s, p, activeCons)
+    assert len(activeCons) == _s.size == y.shape[0]
     #nodes, g = func9(nodes, fo_prev, g, p)
     #y, e = func4(y, e, o, a, fo)
+    if len(nodes) == 0:# can be after removing nodes with nan
+        return _in, g, fo_prev, _s, Solutions, xRecord, r41, r40
     
-
+#    print('asdf')
     if p.solver.dataHandling == 'raw':
-            
         if not isSNLE:
-            tmp = o.copy()
-            tmp[tmp > fo_prev] = -inf
-            M = atleast_1d(nanmax(tmp, 1))
-            for i, node in enumerate(nodes):
-                node.th_key = M[i]
-                node.fo = fo_prev       
+            if 0:
+                tmp = o.copy()
+                tmp[tmp > fo_prev] = -inf
+                M = atleast_1d(nanmax(tmp, 1))
+                assert len(nodes) == M.size
+                for i, node in enumerate(nodes):
+                    node.th_key = M[i]
+                    node.fo = fo_prev       
+            else:
+#                print('asdf')
+                tmp = o.copy()
                 
-        if 0 or p.__isNoMoreThanBoxBounded__():#nlhc is not None and not isSNLE:
-            for node in nodes: node.tnlhf = node.nlhf 
+                # TODO: check is it ever trigger in current interalg implementation
+                tmp[tmp > fo_prev] = 1e300#-inf
+                
+                # TODO: mb change a & o for this ind to 1e300 as well?
+                half_width = tmp.shape[1]//2
+                o_l,  o_u = tmp[:, :half_width], o[:, half_width:]
+                tmp = where((o_l < o_u) | isnan(o_u), o_l, o_u)
+                
+                M = atleast_1d(nanmax(tmp, 1))
+                assert len(nodes) == M.size
+#                ind_exclude = []
+                for i, node in enumerate(nodes):
+#                    if M[i] == 1e300:
+#                        ind_exclude.append(i)
+#                        continue
+                    node.th_key = M[i]
+                    node.fo = fo_prev
+#                    print(p.iter,'th_key:', M[i], 'fo_prev:', fo_prev)
+
+#                for i in ind_exclude[::-1]: 
+#                    nodes.pop(i)
+
+
+        # TODO: MB HANDLE IT INSIDE func11() FUNC
+        if nlhc is None:#TODO: mb rework/improvement for SNLE 
+            for node in nodes: 
+                node.tnlh_fixed = node.nlh_obj_fixed 
         else:
-            for node in nodes: node.tnlhf = node.nlhc + node.nlhf 
+            for node in nodes: 
+#                print('!', p.iter, node.nlh_obj_fixed ,  node.nlhc )
+                node.tnlh_fixed = node.nlh_obj_fixed + node.nlhc 
+                
             
-        an = nodes + _in#hstack((nodes, _in))
+        an = nodes + _in
         
-        #tnlh_fixed = vstack([node.tnlhf for node in an])
-        tnlh_fixed_local = vstack([node.nlhf for node in nodes])#tnlh_fixed[:len(nodes)]
+        
+        #tnlh_fixed = vstack([node.tnlh_fixed for node in an])
+        tnlh_fixed = vstack([node.tnlh_fixed for node in nodes])
+        
+        #tnlh_fixed_local = vstack([node.nlh_obj_fixed for node in nodes])#tnlh_fixed[:len(nodes)]
 #        if isSNLE:
 #            tnlh_curr = tnlh_fixed_local#vstack([node.nlhc for node in nodes])
         if 1:
@@ -118,13 +178,18 @@ def r14(p, nlhc, residual, definiteRange, y, e, vv, asdf1, C,
 #            tmp[tmp>fo_prev] = fo_prev
             tmp2 = tmp_u - tmp_l
 #            ind_inf = tmp2==inf
+
             tmp2[tmp2<1e-300] = 1e-300
+#            tmp2[tmp2<p.fTol] = 1e-300
+            
             o_exclude_ind = o > fo_prev
-            if any(o_exclude_ind):
+            if o_exclude_ind.any():
                 tmp2[o_exclude_ind] = nan
                 if not isSNLE:
                     g = PythonMin(g, min(o[o_exclude_ind]))
-            tnlh_curr = tnlh_fixed_local - log2(tmp2)
+            #tnlh_curr = tnlh_fixed_local - log2(tmp2)
+            tnlh_curr = tnlh_fixed - log2(tmp2)
+            
 #            tnlh_curr[ind_inf] = 1e300
         else:
             if isSNLE:
@@ -136,7 +201,9 @@ def r14(p, nlhc, residual, definiteRange, y, e, vv, asdf1, C,
                 tmp2[tmp2<1e-300] = 1e-300
                 tmp2[o > fo_prev] = nan
                 tnlh_curr = tnlh_fixed_local - log2(tmp2)
+                
         tnlh_curr_best = nanmin(tnlh_curr, 1)
+        
         for i, node in enumerate(nodes):
             node.tnlh_curr = tnlh_curr[i]
             node.tnlh_curr_best = tnlh_curr_best[i]
@@ -146,26 +213,43 @@ def r14(p, nlhc, residual, definiteRange, y, e, vv, asdf1, C,
     else:
         tnlh_curr = None
     
+    
+    
     # TODO: don't calculate PointVals for zero-p regions
     PointVals, PointCoords = getr4Values(vv, y, e, tnlh_curr, asdf1, C, p.contol, dataType, p) 
     if PointVals.size != 0:
+#        print('case1')
         xk, Min = r2(PointVals, PointCoords, dataType)
     else: # all points have been removed by func7
+#        print('case2')
         xk = p.xk
         Min = nan
-
+        
     if r40 > Min:
         r40 = Min
         xRecord = xk.copy()# TODO: is copy required?
-    
-    r41 = PythonMin(Min, r41)
+#    print('Min:',Min)
+    r41 = nanmin((Min, r41))
     
     fo = \
     float(0 if isSNLE else PythonMin(r41, r40 - (fTol if maxSolutions == 1 else 0)))
 
+    tmp = array([node.key for node in (nodes + _in)])
+    f_bound_estimation = tmp.min()
+    
+    if isfinite(r40):
+        p.f_bound_distance = r40 - f_bound_estimation
+    if p.goal in ('max', 'maximum'):
+        f_bound_estimation = -f_bound_estimation
+    p.f_bound_estimation = f_bound_estimation
+    
+#    print '!', p.iter, y, e, hasPoint(y, e, [10.,  20.,  10.,  50])
+
+    assert len(activeCons) == _s.size == y.shape[0]
+
     if p.solver.dataHandling == 'raw':
         
-        if fo != fo_prev and not  isSNLE:
+        if fo != fo_prev and not isSNLE:
             fos = array([node.fo for node in an])
             
             #prev
@@ -176,21 +260,27 @@ def r14(p, nlhc, residual, definiteRange, y, e, vv, asdf1, C,
             delta_fos = fos - fo
             ind_update = where(10 * delta_fos > fos - th_keys)[0]
             
-            update_nlh = True if ind_update.size != 0 else False
+            update_nlh = ind_update.size != 0 
 #                  print 'o MB:', float(o_tmp.nbytes) / 1e6
 #                  print 'percent:', 100*float(ind_update.size) / len(an) 
-            if update_nlh:
+
+# !!!!!!!!!!!!!!!!!!!!
+# TODO: check and enable updateNodes()
+# !!!!!!!!!!!!!!!!!!!!
+            #temporary disabled
+            if 1 and update_nlh:
                 #nodesToUpdate = an[ind_update]
                 nodesToUpdate = [an[i] for i in ind_update]
 #                    from time import time
 #                    tt = time()
-                updateNodes(nodesToUpdate, fo)
+                updateNodes(nodesToUpdate, fo, p.fTol)
 #                    if not hasattr(p, 'Time'):
 #                        p.Time = time() - tt
 #                    else:
 #                        p.Time += time() - tt
-                    
+            
             tmp = array([node.key for node in an])
+            #print('iter:', p.iter, 'tmp:', tmp)
             cond = tmp > fo
             r10 = where(cond)[0]
             g = PythonMin([an[i].key for i in r10]+[g])
@@ -207,11 +297,20 @@ def r14(p, nlhc, residual, definiteRange, y, e, vv, asdf1, C,
             #NN = take(NN, ind, axis=0, out=NN[:ind.size])
             NN = NN[ind]
 
+        #
         if 1 or not isSNLE or p.maxSolutions == 1:
-            #pass
             astnlh = argsort(NN)
-#            print(astnlh[:10])
             an = [an[i] for i in astnlh]#an[astnlh]
+        
+
+        
+        
+        
+#            if isPyPy:
+#                an = [an[i] for i in astnlh]#an[astnlh]
+#            else:
+#                # TODO: 
+#                an = an[astnlh]
             
 #        print(an[0].nlhc, an[0].tnlh_curr_best)
         # Changes
@@ -238,12 +337,14 @@ def r14(p, nlhc, residual, definiteRange, y, e, vv, asdf1, C,
                 arr1 = [node.key for node in _in]
                 arr2 = [node.key for node in nodes]
                 r10 = searchsorted(arr1, arr2)
-                an = insert(_in, r10, nodes)
+                an = insert(_in, r10, nodes).tolist()
 #                if p.debug:
 #                    arr = array([node.key for node in an])
 #                    #print arr[0]
 #                    assert all(arr[1:]>= arr[:-1])
-
+    
+    
+    
     if maxSolutions != 1:
         Solutions = r46(PointCoords, PointVals, fTol, varTols, Solutions)
         
@@ -254,17 +355,27 @@ def r14(p, nlhc, residual, definiteRange, y, e, vv, asdf1, C,
             p.msg = 'user-defined maximal number of solutions (p.maxSolutions = %d) has been exeeded' % p.maxSolutions
             return an, g, fo, None, Solutions, xRecord, r41, r40
     
-    #p.iterfcn(xk, Min)
+#    print('r40:',r40)
     p.iterfcn(xRecord, r40)
-    if not isSNLE and isfinite(r41) and len(an):
+
+    # TODO: fix it (rTol)
+    if 0 and not isSNLE and isfinite(r41) and len(an):
         tmp = array([node.key for node in an])
-        cond_exclude = r41 - tmp <= \
-        p.rTol * where(abs(tmp) < abs(r41), abs(tmp), abs(r41))
+        f_r16estim = tmp.min()
+        abs_tmp = abs(tmp)
+        abs_f_r16estim = abs(f_r16estim)
+#        abs_r40 = abs(r40)
+        
+        #prev
+        cond_exclude = tmp - f_r16estim <= \
+        p.rTol * where(abs_tmp < abs_f_r16estim, abs_tmp, abs_f_r16estim)
         g = PythonMin([g] + [an[j].key for j in where(cond_exclude)[0]])
         cond_remain = logical_not(cond_exclude)
         ind = where(cond_remain)[0]
         an = [an[i] for i in ind]
     
+        g = PythonMin([g] + [an[j].key for j in where(cond_exclude)[0]])
+        
     if p.istop != 0: 
         return an, g, fo, None, Solutions, xRecord, r41, r40
     if isSNLE and maxSolutions == 1 and Min <= fTol:
@@ -272,15 +383,13 @@ def r14(p, nlhc, residual, definiteRange, y, e, vv, asdf1, C,
         p.istop, p.msg = 1000, 'required solution has been obtained'
         return an, g, fo, None, Solutions, xRecord, r41, r40
     
-
-    
     an, g = func9(an, fo, g, p)
-        
+
     nn = maxNodes#1 if asdf1.isUncycled and all(isfinite(o)) and p._isOnlyBoxBounded and not p.probType.startswith('MI') else maxNodes
 
     an, g = func5(an, nn, g, p)
     nNodes.append(len(an))
-        
+    
     return an, g, fo, _s, Solutions, xRecord, r41, r40
 
 
@@ -292,7 +401,7 @@ def r46(PointCoords, PointVals, fTol, varTols, Solutions):
     r5 = PointCoords[r5Ind]
     
     for c in r5:
-        if len(solutions) == 0 or not any(all(abs(c - coords) < varTols, 1)): 
+        if len(solutions) == 0 or not any(all(abs(c - coords) < varTols, axis=1)): 
             solutions.append(c)
             #coords = asarray(solutions)
             Solutions.coords = append(Solutions.coords, c.reshape(1, -1), 0)
@@ -300,28 +409,46 @@ def r46(PointCoords, PointVals, fTol, varTols, Solutions):
     return Solutions
 
 
-def r45(y, e, vv, p, asdf1, dataType, r41, nlhc):
-    o, a, definiteRange, exactRange = func82(y, e, vv, asdf1, dataType, p)#, r41)
-    if p.debug and any(a + 1e-15 < o):  
+def r45(y, e, vv, p, asdf1, dataType, r41, nlhc, _s, indT):
+    
+    y, e, o, a, definiteRange, exactRange, _s, indT = \
+    func82(y, e, vv, asdf1, dataType, p, r41, _s, indT)
+
+    if o is None:
+        return y, e, o, a, r41, _s, indT
+    if p.debug and (a + 1e-15 < o).any():  
         p.warn('interval lower bound exceeds upper bound, it seems to be FuncDesigner kernel bug')
     if p.debug and any(logical_xor(isnan(o), isnan(a))):
         p.err('bug in FuncDesigner intervals engine')
     
-    m, n = e.shape
+    n = p.n
+    m = o.size // (2*n)
+#    print('---')
+#    print(len(_s), m, o.shape,y.shape)
     o, a = o.reshape(2*n, m).T, a.reshape(2*n, m).T
+#    print(len(_s), m, o.shape,y.shape)
+    assert m == y.shape[0] == _s.size
     
-    if p.probType not in ('SNLE', 'NLSP') and asdf1.isUncycled and not p.probType.startswith('MI') \
-    and len(p._discreteVarsList)==0 and exactRange:# for SNLE fo = 0
+    
+    # TODO: check is expected speedup triggers
+    
+
+    
+    if p.probType not in ('SNLE', 'NLSP') and asdf1.isUncycled \
+    and p._continuous_obj_dep_variables and exactRange:# for SNLE fo = 0
         # TODO: 
         # handle constraints with restricted domain and matrix definiteRange
         ind = where(definiteRange)[0]
         
         if ind.size != 0:
             o2 = o[ind]
+            # TODO: mb remove "nlhc is None checks", that currently remain for more safety 
+            # and mb further interalg changes
             if nlhc is not None:
                 nlhc = nlhc[ind]
-            # TODO: if o has at least one -inf => prob is unbounded
-            tmp1 = o2[nlhc==0] if nlhc is not None else o2
+            # TODO: if o has at least one -inf => prob is unbounded in the area y <=x <= e (mb inform it with certificate?)
+            # TODO: is nlhc ever equals to None now? Yet it should remain for more safety, mb for future changes
+            tmp1 = o2[nlhc==p.solver._constraintInactiveValue] if nlhc is not None else o2
             if tmp1.size != 0:
                 tmp1 = nanmin(tmp1)
                 
@@ -332,9 +459,9 @@ def r45(y, e, vv, p, asdf1, dataType, r41, nlhc):
                 
                 r41 = nanmin((r41, tmp1)) 
         
-    return o, a, r41
+    return y, e, o, a, r41, _s, indT
 
-def updateNodes(nodesToUpdate, fo):
+def updateNodes(nodesToUpdate, fo, fTol):
     if len(nodesToUpdate) == 0: return
     a_tmp = vstack([node.a for node in nodesToUpdate])
     Tmp = a_tmp
@@ -342,25 +469,34 @@ def updateNodes(nodesToUpdate, fo):
 
     o_tmp = vstack([node.o for node in nodesToUpdate])
     Tmp -= o_tmp
+    
+    #changes
     Tmp[Tmp<1e-300] = 1e-300
-    Tmp[o_tmp>fo] = nan
-    tnlh_all_new =  - log2(Tmp)
+#    Tmp[Tmp<fTol] = 1e-300#fTol
+    
+# doesn't matter now, should be cleared in other part of interalg code      
+#    Tmp[o_tmp>fo] = nan
+    tnlh_full_new =  - log2(Tmp)
     
     del Tmp, a_tmp
     
-    tnlh_all_new += vstack([node.tnlhf for node in nodesToUpdate])#tnlh_fixed[ind_update]
+    tnlh_full_new += vstack([node.tnlh_fixed for node in nodesToUpdate])#tnlh_fixed[ind_update]
+    if nodesToUpdate[0].nlhc is not None:
+        nlhc = vstack([node.nlhc for node in nodesToUpdate])
+        tnlh_full_new += nlhc
     
-    tnlh_curr_best = nanmin(tnlh_all_new, 1)
+    tnlh_curr_best = nanmin(tnlh_full_new, 1)
 
-    o_tmp[o_tmp > fo] = -inf
+# doesn't matter now, should be cleared in other part of interalg code  
+#    o_tmp[o_tmp > fo] = fo + 1.0
     M = atleast_1d(nanmax(o_tmp, 1))
     for j, node in enumerate(nodesToUpdate): 
         node.fo = fo
-        node.tnlh_curr = tnlh_all_new[j]
+        node.tnlh_curr = tnlh_full_new[j]
         node.tnlh_curr_best = tnlh_curr_best[j]
         node.th_key = M[j]
 
-#    return tnlh_all_new, tnlh_curr_best, M
+#    return tnlh_obj_new, tnlh_curr_best, M
 
 
 #from multiprocessing import Pool
